"""
Module for common application functionality for MESH functions
"""
from mesh_aws_client.mesh_mailbox import MeshMailbox, MeshMessage
from mesh_aws_client.mesh_common import MeshMailbox as OldMeshMailbox
from mesh_aws_client.mesh_common import MeshMessage as OldMeshMessage


__all__ = ["MeshMailbox", "MeshMessage", "OldMeshMailbox", "OldMeshMessage"]



"""Common methods and classes used for mesh client"""
from collections import namedtuple
from http import HTTPStatus
import atexit
import json
import os
import tempfile

from mesh_client import MeshClient
from urllib3.exceptions import InsecureRequestWarning
import boto3
import requests

from spine_aws_common.logger import Logger


class SingletonCheckFailure(Exception):
    """Singleton check failed"""

    def __init__(self, msg=None):
        super().__init__()
        self.msg = msg


class AwsFailedToPerformError(Exception):
    """Errors raised by AWS functions"""

    def __init__(self, msg=None):
        super().__init__()
        self.msg = msg


class MeshCommon:
    """Common"""

    MIB = 1024 * 1024
    DEFAULT_CHUNK_SIZE = 20 * MIB

    @staticmethod
    def singleton_check(mailbox, my_step_function_name):
        """Find out whether there is another step function running for my mailbox"""
        sfn_client = boto3.client("stepfunctions", region_name="eu-west-2")
        response = sfn_client.list_state_machines()
        # Get my step function arn
        my_step_function_arn = None
        for step_function in response.get("stateMachines", []):
            if step_function.get("name", "") == my_step_function_name:
                my_step_function_arn = step_function.get("stateMachineArn", None)

        # TODO add this check to tests
        if not my_step_function_arn:
            raise SingletonCheckFailure(
                "No executing step function arn for "
                + f"step_function={my_step_function_name}"
            )

        response = sfn_client.list_executions(
            stateMachineArn=my_step_function_arn,
            statusFilter="RUNNING",
        )
        currently_running_step_funcs = []
        for execution in response["executions"]:
            currently_running_step_funcs.append(execution["executionArn"])

        exec_count = 0
        for execution_arn in currently_running_step_funcs:
            response = sfn_client.describe_execution(executionArn=execution_arn)
            step_function_input = json.loads(response.get("input", "{}"))
            input_mailbox = step_function_input.get("mailbox", None)
            if input_mailbox == mailbox:
                exec_count = exec_count + 1
            if exec_count > 1:
                raise SingletonCheckFailure("Process already running for this mailbox")

        return True

    @staticmethod
    def convert_params_to_dict(params):
        """Convert paramater dict to key:value dict"""
        new_dict = {}
        for entry in params:
            name = entry.get("Name", None)
            if name:
                var_name = os.path.basename(name)
                new_dict[var_name] = entry.get("Value", None)
        return new_dict

    @staticmethod
    def return_failure(log_object, status, logpoint, mailbox, message=""):
        """Return a failure response with retry"""
        log_object.write_log(logpoint, None, {"mailbox": mailbox, "error": message})
        return {
            "statusCode": status,
            "headers": {
                "Content-Type": "application/json",
                "Retry-After": 18000,
            },
            "body": {
                "internal_id": log_object.internal_id,
                "error": message,
            },
        }

    @staticmethod
    def get_ssm_params(path, recursive=False, decryption=True):
        """Get parameters from SSM param store and return as simple dict"""
        # TODO region name fix
        ssm_client = boto3.client("ssm", region_name="eu-west-2")

        params_result = ssm_client.get_parameters_by_path(
            Path=path,
            Recursive=recursive,
            WithDecryption=decryption,
        )
        params = params_result.get("Parameters", {})
        new_params_dict = {}
        for entry in params:
            name = entry.get("Name", None)
            if name:
                var_name = os.path.basename(name)
                new_params_dict[var_name] = entry.get("Value", None)
        return new_params_dict


class ExtendedMeshClient(MeshClient):
    """Extended functionality for lambda send"""


# Named tuple for holding Mesh Message info
MeshMessage = namedtuple(
    "MeshMessage",
    [
        "filename",
        "body",
        "src_mailbox",
        "dest_mailbox",
        "workflow_id",
        "message_id",
    ],
)


class MeshMailbox:  # pylint: disable=too-many-instance-attributes
    """
    Mesh mailbox object, gets parameters from SSM Parameter Store
    DEPRECATED Version
    """

    VERSION = "0.0.1"

    def __init__(self, log_object: Logger, mailbox, environment="default"):
        self.mailbox = mailbox
        self.common_params = None
        self.mailbox_params = None
        self.ca_cert_file = None
        self.client_cert_file = None
        self.client_key_file = None
        self.environment = environment
        self.dest_mailbox = None
        self.workflow_id = None
        self.temp_dir_object = None
        self.log_object = log_object
        self._setup()
        atexit.register(self.clean_up)

    def clean_up(self):
        """Close mesh client at exit"""
        self.mesh_client.close()

    def _setup(self):
        """Get the parameters from SSM paramter store"""
        # TODO refactor
        ssm_client = boto3.client("ssm", region_name="eu-west-2")
        self.log_object.write_log(
            "MESH0001", None, {"mailbox": self.mailbox, "environment": self.environment}
        )
        common_params_result = ssm_client.get_parameters_by_path(
            Path=f"/{self.environment}/mesh", Recursive=False, WithDecryption=True
        )
        self.common_params = MeshCommon.convert_params_to_dict(
            common_params_result.get("Parameters", {})
        )
        mailbox_params_result = ssm_client.get_parameters_by_path(
            Path=f"/{self.environment}/mesh/mailboxes/{self.mailbox}",
            Recursive=False,
            WithDecryption=True,
        )
        self.mailbox_params = MeshCommon.convert_params_to_dict(
            mailbox_params_result.get("Parameters", {})
        )
        self._write_certs_to_files()

        maybe_verify = bool(self.common_params.get("MESH_VERIFY_SSL") == "True")

        if not maybe_verify:
            requests.urllib3.disable_warnings(InsecureRequestWarning)
        self.mesh_client = ExtendedMeshClient(
            self.common_params["MESH_URL"],
            self.mailbox,
            self.mailbox_params["MAILBOX_PASSWORD"],
            shared_key=self.common_params["MESH_SHARED_KEY"].encode("utf8"),
            cert=(self.client_cert_file.name, self.client_key_file.name),
            verify=self.ca_cert_file.name if maybe_verify else None,
            max_chunk_size=MeshCommon.DEFAULT_CHUNK_SIZE,
        )

    def _write_certs_to_files(self):
        """Write the certificates to a local file"""
        # pylint: disable=consider-using-with
        self.temp_dir_object = tempfile.TemporaryDirectory()
        temp_dir = self.temp_dir_object.name

        # store as temporary files for the mesh client
        self.client_cert_file = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False)
        client_cert = self.common_params["MESH_CLIENT_CERT"]
        self.client_cert_file.write(client_cert.encode("utf-8"))
        self.client_cert_file.seek(0)

        self.client_key_file = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False)
        client_key = self.common_params["MESH_CLIENT_KEY"]
        self.client_key_file.write(client_key.encode("utf-8"))
        self.client_key_file.seek(0)

        self.ca_cert_file = None
        if self.common_params.get("MESH_VERIFY_SSL", False) == "True":
            self.ca_cert_file = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False)
            ca_cert = self.common_params["MESH_CA_CERT"]
            self.ca_cert_file.write(ca_cert.encode("utf-8"))
            self.ca_cert_file.seek(0)
        # pylint: enable=consider-using-with

    def set_destination_and_workflow(self, dest_mailbox, workflow_id):
        """Set destination mailbox and workflow_id"""
        self.dest_mailbox = dest_mailbox
        self.workflow_id = workflow_id

    def authenticate(self):
        """Authenticate to MESH mailbox"""
        return self.mesh_client.handshake()

    def send_chunk(
        self,
        mesh_message_object,
        chunk=False,
        chunk_size=MeshCommon.DEFAULT_CHUNK_SIZE,
        chunk_num=1,
    ):
        """Send a chunk"""
        # override mailbox dest_mailbox if provided in message_object
        if mesh_message_object.dest_mailbox:
            dest_mailbox = mesh_message_object.dest_mailbox
        else:
            dest_mailbox = self.mailbox
        # override mailbox workflow_id if provided in message_object
        if mesh_message_object.workflow_id:
            workflow_id = mesh_message_object.workflow_id
        else:
            workflow_id = self.mailbox.workflow_id
        if not chunk:
            message_id = self.mesh_client.send_message(
                dest_mailbox,
                workflow_id=workflow_id,
                filename=mesh_message_object.filename,
                data=mesh_message_object.body,
            )
            mesh_message_object = mesh_message_object._replace(message_id=message_id)

        else:
            # TODO chunking is more interesting, as mesh_client doesn't have a
            # function for sending one chunk only
            offset = chunk_size * chunk_num
            print(f"Calculated offset is {offset}")
            print("CHUNKING ON SEND IS NOT IMPLEMENTED YET")
            return (HTTPStatus.NOT_IMPLEMENTED.value, None)
        return (HTTPStatus.OK.value, mesh_message_object)

    def get_chunk(
        self,
        message_id,
        chunk=False,
        chunk_size=MeshCommon.DEFAULT_CHUNK_SIZE,
        chunk_num=1,
    ):
        """Get a chunk"""
        if not chunk:
            message_object = self.mesh_client.retrieve_message(message_id)
            # for (header, value) in message_object.mex_headers():
            #     print(f"{header}={value}")
            filename = message_object.mex_header("filename")
            if not filename:
                filename = message_id
            workflow_id = message_object.mex_header(
                "workflowid", default="Not Provided"
            )
            src_mailbox = message_object.mex_header("from", default="Not Provided")
            # TODO get other info available - compression etc
            return_message_object = MeshMessage(
                filename=filename,
                message_id=message_id,
                body=message_object.read(),
                dest_mailbox=self.mailbox,
                src_mailbox=src_mailbox,
                workflow_id=workflow_id,
            )
            return (HTTPStatus.OK.value, return_message_object)

        # chunked
        offset = chunk_size * chunk_num
        print(f"Calculated offset is {offset}")
        print("CHUNKING ON RECEIVE IS NOT IMPLEMENTED YET")
        return (HTTPStatus.NOT_IMPLEMENTED.value, None)



"""
Base DynamoDB streams Lambda application
"""
from aws_lambda_powertools.utilities.data_classes import DynamoDBStreamEvent
from spine_aws_common.batch_application import BatchApplication


class DynamoDBStreamsApplication(BatchApplication):
    """
    Base class for SQS Lambda applications
    """

    EVENT_TYPE = DynamoDBStreamEvent

    def _get_internal_id_from_record(self, record):
        """
        Always create new internalID for DynamoDB Streams event
        """
        return self._create_new_internal_id()



"""
Module for MESH API functionality for step functions
"""
from http import HTTPStatus
import os
import json

import boto3
from botocore.errorfactory import ClientError

from spine_aws_common import LambdaApplication
from mesh_aws_client.mesh_common import AwsFailedToPerformError, MeshCommon
from mesh_aws_client.mesh_mailbox import MeshMailbox


class MeshFetchMessageChunkApplication(
    LambdaApplication
):  # pylint: disable=too-many-instance-attributes
    """
    MESH API Lambda for sending a message
    """

    MEBIBYTE = 1024 * 1024
    DEFAULT_BUFFER_SIZE = 5 * MEBIBYTE

    def __init__(self, additional_log_config=None, load_ssm_params=False):
        """
        Init variables
        """
        super().__init__(additional_log_config, load_ssm_params)
        self.mailbox = None
        self.old_mailbox = None  # TODO remove
        self.input = {}
        self.environment = os.environ.get("Environment", "default")
        self.chunk_size = os.environ.get("CHUNK_SIZE", MeshCommon.DEFAULT_CHUNK_SIZE)
        self.http_response = None
        self.response = {}
        self.internal_id = None
        self.aws_upload_id = None
        self.aws_current_part_id = 1
        self.aws_part_etags = []
        self.chunked = False
        self.current_chunk = 1
        self.message_id = None
        self.region = os.environ.get("AWS_REGION", "eu-west-2")
        self.s3_client = None
        self.s3_bucket = ""
        self.s3_key = ""

    def initialise(self):
        """decode input event"""
        self.input = self.event.get("body")
        self.internal_id = self.input.get("internal_id", "Not Provided")
        self.aws_upload_id = self.input.get("aws_upload_id", None)
        self.aws_current_part_id = self.input.get("aws_current_part_id", 1)
        self.aws_part_etags = self.input.get("aws_part_etags", [])
        self.chunked = self.input.get("chunked")
        self.current_chunk = self.input.get("chunk_num", 1)
        self.message_id = self.input["message_id"]
        self.response = self.event.raw_event

    def _setup_mailbox(self):
        self.mailbox = MeshMailbox(
            self.log_object, self.input["dest_mailbox"], self.environment
        )

    def _get_filename(self):
        file_name = self.http_response.headers.get("Mex-Filename", "")
        if len(file_name) == 0:
            if self.http_response.headers.get("Mex-Messagetype") == "REPORT":
                file_name = self.message_id + ".ctl"
            else:
                file_name = self.message_id + ".dat"
        return file_name

    def _get_aws_bucket_and_key(self):
        self.s3_bucket = self.mailbox.params["INBOUND_BUCKET"]
        s3_folder = self.mailbox.params.get("INBOUND_FOLDER", "")
        if len(s3_folder) > 0:
            s3_folder += "/"
        file_name = self._get_filename()
        self.s3_key = s3_folder + (
            file_name if len(file_name) > 0 else self.message_id + ".dat"
        )

    def _is_last_part(self, part_num):
        content_length = int(self.http_response.headers["Content-Length"])
        last_part = part_num > content_length / self.DEFAULT_BUFFER_SIZE
        return last_part

    def _is_last_chunk(self, chunk_num):
        chunk_range = self.http_response.headers.get("Mex-Chunk-Range", "1:1")
        number_of_chunks = int(chunk_range[2:])
        last_chunk = chunk_num == number_of_chunks
        return last_chunk

    def _upload_part_to_s3(self, buffer, last_chunk, last_part):
        """Upload a part to S3 and check response"""

        # check if part_overflow_{message_id}.tmp exists and pre-pend to buffer
        filename = f"part_overflow_{self.message_id}.tmp"
        s3_tempfile_key = os.path.basename(self.s3_key) + filename
        try:
            s3_response = self.s3_client.get_object(
                Bucket=self.s3_bucket, Key=s3_tempfile_key
            )
            if s3_response["ResponseMetadata"]["HTTPStatusCode"] == HTTPStatus.OK.value:
                pre_buffer = s3_response["Body"].read()
                buffer = pre_buffer + buffer
                self.log_object.write_log(
                    "MESHFETCH0002b",
                    None,
                    {
                        "aws_part_size": len(pre_buffer),
                        "aws_upload_id": self.aws_upload_id,
                    },
                )
            self.s3_client.delete_object(
                Bucket=self.s3_bucket,
                Key=s3_tempfile_key,
                BypassGovernanceRetention=True,
            )
        except ClientError:
            # Not found
            pass

        if not (last_chunk and last_part):  # not on last chunk
            if len(buffer) < 5 * self.MEBIBYTE:
                # write to s3 if buffer is less than 5 MebiBytes and last part of chunk
                if last_part:
                    self.s3_client.put_object(
                        Bucket=self.s3_bucket,
                        Key=s3_tempfile_key,
                        Body=buffer,
                    )
                self.log_object.write_log(
                    "MESHFETCH0002a",
                    None,
                    {
                        "aws_part_size": len(buffer),
                        "aws_upload_id": self.aws_upload_id,
                    },
                )
                return None

        response = self.s3_client.upload_part(
            Body=buffer,
            Bucket=self.s3_bucket,
            Key=self.s3_key,
            PartNumber=self.aws_current_part_id,
            ContentLength=len(buffer),
            UploadId=self.aws_upload_id,
        )
        # check return code
        if response["ResponseMetadata"].get("HTTPStatusCode") != HTTPStatus.OK.value:
            self.response.update({"statusCode": HTTPStatus.INTERNAL_SERVER_ERROR.value})
            # logpoint
            raise AwsFailedToPerformError(
                f'Failed to partial upload key="{self.s3_key}" into'
                + f' bucket="{self.s3_bucket}"'
                + f' part_id="{self.aws_current_part_id}" to save from MESH'
            )
        etag = response["ETag"]
        self.aws_part_etags.append(
            {
                "ETag": etag,
                "PartNumber": self.aws_current_part_id,
            }
        )
        self.aws_current_part_id += 1
        self.log_object.write_log(
            "MESHFETCH0002",
            None,
            {
                "aws_part_id": self.aws_current_part_id,
                "aws_part_size": len(buffer),
                "aws_upload_id": self.aws_upload_id,
                "etag": etag,
            },
        )
        return etag

    def _create_multipart_upload(self):
        """Create an S3 multipart upload"""
        multipart_upload = self.s3_client.create_multipart_upload(
            Bucket=self.s3_bucket,
            Key=self.s3_key,
        )
        self.aws_upload_id = multipart_upload["UploadId"]
        # check return code
        if (
            multipart_upload["ResponseMetadata"].get("HTTPStatusCode")
            != HTTPStatus.OK.value
        ):
            self.response.update({"statusCode": HTTPStatus.INTERNAL_SERVER_ERROR.value})
            # logpoint
            raise AwsFailedToPerformError(
                f'Failed to create multi-part upload key="{self.s3_key}" into '
                + f'bucket="{self.s3_bucket}"'
                + f' aws_upload_id="{self.aws_upload_id}" to save from MESH'
            )

    def _finish_multipart_upload(self):
        """Complete the s3 multipart upload"""
        response = self.s3_client.complete_multipart_upload(
            Bucket=self.s3_bucket,
            Key=self.s3_key,
            UploadId=self.aws_upload_id,
            MultipartUpload={"Parts": self.aws_part_etags},
        )
        # check return code
        if response["ResponseMetadata"].get("HTTPStatusCode") != HTTPStatus.OK.value:
            self.response.update({"statusCode": HTTPStatus.INTERNAL_SERVER_ERROR.value})
            # logpoint
            raise AwsFailedToPerformError(
                f'Failed to complete multi-part upload key="{self.s3_key}"'
                + f'into bucket="{self.s3_bucket}"'
                + f' aws_upload_id="{self.aws_upload_id}" to save from MESH'
            )

    def start(self):
        """
        Main body of lambda function
        """
        self.s3_client = boto3.client("s3", region_name=self.region)

        self.log_object.internal_id = self.internal_id
        self._setup_mailbox()

        self.log_object.write_log(
            "MESHFETCH0001",
            None,
            {
                "message_id": self.message_id,
            },
        )

        # get stream for this chunk
        self.http_response = self.mailbox.get_chunk(
            self.message_id, chunk_num=self.current_chunk
        )
        self.http_response.raise_for_status()
        self.chunked = (
            self.http_response.status_code == HTTPStatus.PARTIAL_CONTENT.value
        )

        # get s3 bucket and key
        self._get_aws_bucket_and_key()

        if self.current_chunk == 1:
            # create multipart upload even if only one chunk
            self._create_multipart_upload()

        parts_read = 0
        part_buffer = b""

        is_report = self.http_response.headers.get("Mex-Messagetype") == "REPORT"
        if is_report:
            buffer = json.dumps(dict(self.http_response.headers))
            etag = self._upload_part_to_s3(buffer, True, True)

        if self.http_response.headers.get("Mex-Messagetype") == "DATA":
            # read bytes into buffer
            for buffer in self.http_response.iter_content(
                chunk_size=self.DEFAULT_BUFFER_SIZE
            ):
                parts_read += 1
                last_part = self._is_last_part(parts_read)
                last_chunk = self._is_last_chunk(self.current_chunk)
                etag = self._upload_part_to_s3(
                    part_buffer + buffer, last_chunk, last_part
                )
                if etag:
                    part_buffer = b""
                else:
                    part_buffer = buffer

        is_finished = not self.chunked
        if is_finished:
            self._finish_multipart_upload()
            self.log_object.write_log(
                "MESHFETCH0004", None, {"message_id": self.message_id}
            )
            self.mailbox.acknowledge_message(self.message_id)
        else:
            self.log_object.write_log(
                "MESHFETCH0003",
                None,
                {"chunk": self.current_chunk, "message_id": self.message_id},
            )
            self.current_chunk += 1

        # update event to send as response
        self.response.update({"statusCode": self.http_response.status_code})
        self.response["body"].update(
            {
                "complete": is_finished,
                "chunk_num": self.current_chunk,
                "aws_upload_id": self.aws_upload_id,
                "aws_current_part_id": self.aws_current_part_id,
                "aws_part_etags": self.aws_part_etags,
                "internal_id": self.internal_id,
                "file_name": self._get_filename(),
            }
        )


# create instance of class in global space
# this ensures initial setup of logging/config is only done on cold start
app = MeshFetchMessageChunkApplication()


def lambda_handler(event, context):
    """Standard lambda_handler"""
    return app.main(event, context)



"""
Module for MESH API functionality for step functions
"""
import json
import gzip
from http import HTTPStatus
import os

import boto3

from spine_aws_common import LambdaApplication
from mesh_aws_client.mesh_mailbox import MeshMailbox, MeshMessage
from mesh_aws_client.mesh_common import MeshCommon


class MeshSendMessageChunkApplication(
    LambdaApplication
):  # pylint: disable=too-many-instance-attributes
    """
    MESH API Lambda for sending a message / message chunk
    """

    MEBIBYTE = 1024 * 1024
    DEFAULT_BUFFER_SIZE = 20 * MEBIBYTE

    def __init__(self, additional_log_config=None, load_ssm_params=False):
        """
        Init variables
        """
        super().__init__(additional_log_config, load_ssm_params)
        self.mailbox = None
        self.input = {}
        self.body = None
        self.environment = os.environ.get("Environment", "default")
        self.chunked = False
        self.current_byte = 0
        self.current_chunk = 1
        self.chunk_size = 0
        self.compression_ratio = 1
        self.will_compress = False
        self.s3_client = None
        self.bucket = ""
        self.key = ""
        self.buffer_size = self.DEFAULT_BUFFER_SIZE
        self.file_size = 0

    def initialise(self):
        """Setup class variables"""
        self.input = self.event.get("body")
        self.response = self.event.raw_event
        self.current_byte = self.input.get("current_byte_position", 0)
        self.current_chunk = self.input.get("chunk_number", 1)
        self.chunk_size = self.input.get("chunk_size", MeshCommon.DEFAULT_CHUNK_SIZE)
        self.chunked = self.input.get("chunked", False)
        self.compression_ratio = self.input.get("compress_ratio", 1)
        self.will_compress = self.input.get("will_compress", False)
        self.s3_client = boto3.client("s3")
        self.bucket = self.input["bucket"]
        self.key = self.input["key"]
        return super().initialise()

    def _get_file_from_s3(self):
        """Get a file or chunk of a file from S3"""
        start_byte = self.current_byte
        end_byte = start_byte + (self.chunk_size * self.compression_ratio)
        if end_byte > self.file_size:
            end_byte = self.file_size
        while self.current_byte < end_byte:
            bytes_to_end = end_byte - self.current_byte
            if bytes_to_end > self.buffer_size:
                range_spec = (
                    f"bytes={self.current_byte}-"
                    + f"{self.current_byte + self.buffer_size - 1}"
                )
                self.current_byte = self.current_byte + self.buffer_size
            else:
                range_spec = f"bytes={self.current_byte}-{end_byte-1}"
                self.current_byte = end_byte

            response = self.s3_client.get_object(
                Bucket=self.bucket, Key=self.key, Range=range_spec
            )

            body = response.get("Body", None)
            if body:
                file_content = body.read()
                if len(file_content) == 0:
                    file_content = None
                self.log_object.write_log(
                    "MESHSEND0006",
                    None,
                    {
                        "file": self.key,
                        "bucket": self.bucket,
                        "num_bytes": len(file_content),
                        "byte_range": range_spec,
                    },
                )
            else:
                file_content = None

            if self.will_compress:
                compressed_bytes = gzip.compress(file_content)
                yield compressed_bytes
            else:
                yield file_content

    def start(self):
        """Main body of lambda"""

        is_finished = self.input.get("complete", False)
        if is_finished:
            # TODO log error
            self.response.update({"statusCode": HTTPStatus.INTERNAL_SERVER_ERROR.value})
            raise SystemError("Already completed upload to MESH")

        total_chunks = self.input.get("total_chunks", 1)
        message_id = self.input.get("message_id", None)

        file_response = self.s3_client.head_object(Bucket=self.bucket, Key=self.key)
        self.file_size = file_response["ContentLength"]
        self.log_object.write_log(
            "MESHSEND0005",
            None,
            {
                "file": self.key,
                "bucket": self.bucket,
                "chunk_num": self.current_chunk,
                "max_chunk": total_chunks,
            },
        )

        self.mailbox = MeshMailbox(
            self.log_object, self.input["src_mailbox"], self.environment
        )
        self.mailbox.set_destination_and_workflow(
            self.input["dest_mailbox"], self.input["workflow_id"]
        )

        message_object = MeshMessage(
            file_name=os.path.basename(self.key),
            data=self._get_file_from_s3(),
            message_id=message_id,
            dest_mailbox=self.mailbox.dest_mailbox,
            src_mailbox=self.mailbox.mailbox,
            workflow_id=self.mailbox.workflow_id,
            will_compress=self.will_compress,
        )
        if self.file_size > 0:
            response = self.mailbox.send_chunk(
                mesh_message_object=message_object,
                number_of_chunks=total_chunks,
                chunk_num=self.current_chunk,
            )
            status_code = response.status_code
            message_id = json.loads(response.text)["messageID"]
            status_code = HTTPStatus.OK.value
        else:
            status_code = HTTPStatus.NOT_FOUND.value
            self.response.update({"statusCode": status_code})
            raise FileNotFoundError

        is_finished = self.current_chunk >= total_chunks if self.chunked else True
        if self.chunked and not is_finished:
            self.current_chunk += 1

        # update input event to send as response
        self.response.update({"statusCode": status_code})
        self.response["body"].update(
            {
                "complete": is_finished,
                "message_id": message_id,
                "chunk_number": self.current_chunk,
                "current_byte_position": self.current_byte,
            }
        )


# create instance of class in global space
# this ensures initial setup of logging/config is only done on cold start
app = MeshSendMessageChunkApplication()


def lambda_handler(event, context):
    """Standard lambda_handler"""
    return app.main(event, context)



"""
Base SES Lambda application
"""
from aws_lambda_powertools.utilities.data_classes import SESEvent
from spine_aws_common.batch_application import BatchApplication


class SESApplication(BatchApplication):
    """
    Base class for SES Lambda applications
    """

    EVENT_TYPE = SESEvent



"""
Module for common application functionality for Lambda functions
"""
import os
import sys
from datetime import datetime, timezone
import uuid
from abc import abstractmethod
from aws_lambda_powertools.utilities import parameters
from aws_lambda_powertools.utilities.data_classes.common import DictWrapper
from aws_lambda_powertools.utilities.typing.lambda_context import LambdaContext
from spine_aws_common.logger import Logger, configure_logging_adapter
from spine_aws_common.utilities import StopWatch

DELIMITER = "_"


# pylint: disable=too-many-instance-attributes
class LambdaApplication:
    """
    Base class for Lambda applications
    """

    # Base class will always return event in same format
    EVENT_TYPE = DictWrapper

    def __init__(self, additional_log_config=None, load_ssm_params=False):
        self.context = None
        self.event = None
        self.sync_timer = None

        self.system_config = self._load_system_config(load_ssm_params=load_ssm_params)

        self.log_object = self.get_logger(additional_log_config=additional_log_config)
        configure_logging_adapter(self.log_object)

        self._log_coldstart()

        self.response = None

    def main(self, event, context):
        """
        Common entry point behaviour
        """
        self.response = {"message": "Lambda application stopped"}
        try:
            self.sync_timer = StopWatch()
            self.sync_timer.start_the_clock()
            self.context = context
            self.event = self.process_event(event)
            self.log_object.set_internal_id(self._get_internal_id())

            self._log_start()

            self.initialise()

            self.start()

            self._log_end()

        except InitialisationError as e:
            if self.log_object is None:
                print(e)
            else:
                self.log_object.write_log("LAMBDAINIT001", None, {"message": e})
            raise e
        except Exception as e:  # pylint:disable=broad-except
            if self.log_object is None:
                print(e)
            else:
                self.log_object.write_log(
                    "LAMBDA9999", sys.exc_info(), {"error": str(e)}
                )
            raise e

        return self.response

    def get_logger(self, additional_log_config=None):
        """
        Gets the default application logger. This may be overridden
        if a custom logger is required.
        """

        logger = Logger(
            process_name=self.system_config.get("AWS_LAMBDA_FUNCTION_NAME", "None"),
            additional_log_config=additional_log_config,
        )
        return logger

    def process_event(self, event):
        """
        Processes event object passed in by Lambda service
        Can be overridden to customise event parsing
        """
        return self.EVENT_TYPE(event)

    def initialise(self):
        """
        Application initialisation
        """

    @abstractmethod
    def start(self):
        """
        Start the application
        """

    def _get_internal_id(self):
        """
        Get internalID from the event
        """
        return self.event.get("internal_id") or self._create_new_internal_id()

    @staticmethod
    def _create_new_internal_id():
        """
        Creates an internalID
        """
        process_start_time = datetime.now(timezone.utc)
        internal_id = process_start_time.strftime("%Y%m%d%H%M%S%f")
        internal_id += DELIMITER + str(uuid.uuid4())[:6].upper()
        return internal_id

    @staticmethod
    def _load_system_config(load_ssm_params: bool):
        """
        Load common system configuration from Lambda ENV vars
        """
        try:
            env = os.environ.copy()
            process_name = env.get("AWS_LAMBDA_FUNCTION_NAME", "None")
            if load_ssm_params:
                config = parameters.get_parameters(
                    f"/{process_name}", force_fetch=True, decrypt=True
                )
                config.update(env)
                return config
            return env
        except Exception as e:
            raise InitialisationError(e) from e

    def _log_coldstart(self):
        log_params = {
            "aws_region": self.system_config.get("AWS_REGION"),
            "aws_execution_env": self.system_config.get("AWS_EXECUTION_ENV"),
            "function_name": self.system_config.get("AWS_LAMBDA_FUNCTION_NAME", "None"),
            "function_memory_size": self.system_config.get(
                "AWS_LAMBDA_FUNCTION_MEMORY_SIZE"
            ),
            "function_version": self.system_config.get("AWS_LAMBDA_FUNCTION_VERSION"),
        }
        self.log_object.write_log("LAMBDA0001", None, log_params)

    def _log_start(self):
        log_params = {
            "aws_request_id": self._get_aws_request_id(),
        }
        self.log_object.write_log("LAMBDA0002", None, log_params)

    def _log_end(self):
        log_params = {
            "duration": self.sync_timer.stop_the_clock(),
            "aws_request_id": self._get_aws_request_id(),
        }
        self.log_object.write_log("LAMBDA0003", None, log_params)

    def _get_aws_request_id(self):
        """Get the request id"""
        if isinstance(self.context, LambdaContext) and getattr(
            self.context, "aws_request_id", None
        ):
            return self.context.aws_request_id
        if isinstance(self.context, dict):
            return self.context.get("aws_request_id", "unknown")
        return "unknown"


def overrides(base_class):
    """
    Decorator used to specify that a method overrides a base class method
    """

    def decorate(method):
        """
        Override assertion
        """
        assert method.__name__ in dir(base_class)
        return method

    return decorate


class InitialisationError(Exception):
    """
    Application initialisation error
    """

    def __init__(self, msg=None):
        super().__init__()
        self.msg = msg
